def desolve(de,vars):
    """
    Solves a 1st or 2nd order ODE via maxima.

    INPUT: 
        de    -- a string representing the ODE 
                 (eg, de = "diff(f(x),x,2)=diff(f(x),x)+sin(x)")
        vars  -- a list of strings representing the variables
                 (eg, vars = ["x","f"])

    EXAMPLES:
        sage: desolve("x^2*diff(y,x) + 3*y*x = sin(x)/x",["x","y"])
        (%c - cos(x))/x^3

    AUTHOR: David Joyner (1-2006)
    """
    maxima("depends("+vars[1]+","+vars[0]+");")
    maxima("de:"+de+";")
    cmd = "ode2("+de+","+vars[1]+","+vars[0]+");"
    print maxima(cmd)
    #maxima.eval(cmd)

def desolve_laplace(de,vars,ics=None):
    """
    Solves an ODE using laplace transforms via maxima.

    INPUT: 
        de    -- a string representing the ODE 
                 (eg, de = "diff(f(x),x,2)=diff(f(x),x)+sin(x)")
        vars  -- a list of strings representing the variables
                 (eg, vars = ["x","f"])
        ics   -- a list of numbers representing initial conditions,
                 with symbols allowed which are represented by strings
                 (eg, f(0)=1, f'(0)=2 is ics = [0,1,2])
           
    EXAMPLES:
        sage: desolve_laplace("diff(f(x),x,2)=2*diff(f(x),x)-f(x)",["x","f"])
        x*%e^x*(at('diff(f(x),x,1),x = 0)) - f(0)*x*%e^x + f(0)*%e^x
        sage: desolve_laplace("diff(f(x),x,2)=2*diff(f(x),x)-f(x)",["x","f"],[0,1,2])
        x*%e^x + %e^x

    WARNING: 
        The second equation sets the values of f(0) and f'(0) in maxima, so
        subsequent ODEs involving these variables will have these initial conditions
        automatically imposed. 

    AUTHOR: David Joyner (1-2006)
    """
    maxima("de:"+de+";")
    if ics!=None:
        d = len(ics)
        for i in range(0,d-1):
            ic = "atvalue(diff("+vars[1]+"("+vars[0]+"),"+str(vars[0])+","+str(i)+"),"+str(vars[0])+"="+str(ics[0])+","+str(ics[1+i])+")"
            maxima(ic)
            #print i,ic
    cmd = "desolve("+de+","+vars[1]+"("+vars[0]+"));"
    print maxima(cmd)
    #maxima.eval(cmd)


