#############################################################################
##
#W  semirel.gd                  GAP library                   James D Mitchell
##
#H  @(#)$Id: semirel.gd,v 4.20.2.1 2006/02/27 12:08:28 sal Exp $
##
#Y  Copyright (C)  1997,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
#Y  (C) 1998 School Math and Comp. Sci., University of St.  Andrews, Scotland
#Y  Copyright (C) 2002 The GAP Group
##
##  This file contains the declarations for equivalence relations on 
##  semigroups. Of particular interest are Green's relations, 
##  congruences, and Rees congruences.
##
Revision.semirel_gd :=
    "@(#)$Id: semirel.gd,v 4.20.2.1 2006/02/27 12:08:28 sal Exp $";

#######################
#######################

DeclareCategory("IsGreensRelation", IsBinaryRelation);
DeclareCategory("IsGreensRRelation", IsGreensRelation);
DeclareCategory("IsGreensLRelation", IsGreensRelation);
DeclareCategory("IsGreensJRelation", IsGreensRelation);
DeclareCategory( "IsGreensHRelation", IsGreensRelation);
DeclareCategory( "IsGreensDRelation", IsGreensRelation);

DeclareProperty("IsFiniteSemigroupGreensRelation", IsGreensRelation);


#######################
#######################

DeclareAttribute("GreensRRelation", IsSemigroup);
DeclareAttribute("GreensLRelation", IsSemigroup);
DeclareAttribute("GreensJRelation", IsSemigroup);
DeclareAttribute("GreensDRelation", IsSemigroup);
DeclareAttribute("GreensHRelation", IsSemigroup);

#############################################################################
##
#O  GreensRClassOfElement(<S>, <a>)
#O  GreensLClassOfElement(<S>, <a>)
#O  GreensDClassOfElement(<S>, <a>)
#O  GreensJClassOfElement(<S>, <a>)
#O  GreensHClassOfElement(<S>, <a>)
##
##  Creates the <> class of the element <a> in the semigroup <S>
##  where <> is one of L, R, D, J or H.

DeclareOperation("GreensRClassOfElement", [IsSemigroup, IsObject]);
DeclareOperation("GreensLClassOfElement", [IsSemigroup, IsObject]);
DeclareOperation("GreensDClassOfElement", [IsSemigroup, IsObject]);
DeclareOperation("GreensJClassOfElement", [IsSemigroup, IsObject]);
DeclareOperation("GreensHClassOfElement", [IsSemigroup, IsObject]);

#######################
#######################

DeclareOperation("FroidurePinSimpleAlg", [IsMonoid and HasIsFinite and IsFinite]);
DeclareOperation("FroidurePinExtendedAlg", [IsSemigroup and HasIsFinite and IsFinite]);

DeclareAttribute("AssociatedConcreteSemigroup", IsFpSemigroup);
DeclareAttribute("AssociatedFpSemigroup", IsSemigroup);

DeclareSynonymAttr("LeftCayleyGraphSemigroup", CayleyGraphDualSemigroup);
DeclareSynonymAttr("RightCayleyGraphSemigroup", CayleyGraphSemigroup);

DeclareProperty("IsGreensClass", IsEquivalenceClass);
DeclareProperty("IsGreensRClass", IsEquivalenceClass);
DeclareProperty("IsGreensLClass", IsEquivalenceClass);
DeclareProperty("IsGreensJClass", IsEquivalenceClass);
DeclareProperty("IsGreensHClass", IsEquivalenceClass);
DeclareProperty("IsGreensDClass", IsEquivalenceClass);

InstallTrueMethod(IsGreensClass, IsGreensRClass);
InstallTrueMethod(IsGreensClass, IsGreensLClass);
InstallTrueMethod(IsGreensClass, IsGreensJClass);
InstallTrueMethod(IsGreensClass, IsGreensHClass);
InstallTrueMethod(IsGreensClass, IsGreensDClass);


#############################################################################
##
#A  AssociatedSemigroup(<greens-class>) . . . . . . . . .   for Green's class
##
##  A greens class needs what semigroup it is associated with
##

DeclareSynonymAttr("AssociatedSemigroup", ParentAttr);

#############################################################################
##
#A  GreensRClasses(<semigroup>)
#A  GreensLClasses(<semigroup>)
#A  GreensJClasses(<semigroup>)
#A  GreensDClasses(<semigroup>)
#A  GreensHClasses(<semigroup>)
##
##  return the <R>, <L>, <J>, <H>, or <D> Green's classes, respectively for
##  semigroup <semigroup>. EquivlanceClasses for a Green's relation lead to one of 
##  these functions.
##  

DeclareAttribute("GreensRClasses", IsSemigroup);
DeclareAttribute("GreensLClasses", IsSemigroup);
DeclareAttribute("GreensJClasses", IsSemigroup);
DeclareAttribute("GreensDClasses", IsSemigroup);
DeclareAttribute("GreensHClasses", IsSemigroup);

DeclareAttribute("GreensHClasses", IsGreensClass);
DeclareAttribute("GreensRClasses", IsGreensDClass);
DeclareAttribute("GreensLClasses", IsGreensDClass);

#############################################################################
## 
#O  IsGreensLessThanOrEqual( <C1>, <C2> )
##
##  returns `true' if the greens class <C1> is less than or equal to <C2> 
##  under the respective ordering (as defined above), and `false' otherwise. 
##
##  Only defined for R, L and J classes.
##
DeclareOperation("IsGreensLessThanOrEqual", [IsGreensClass, IsGreensClass]);

#############################################################################
## 
#A  RClassOfHClass( <H> )
#A  LClassOfHClass( <H> )
##
##  are attributes reflecting the natural ordering over the various Green's
##  classes. `RClassOfHClass' and `LClassOfHClass' return the <R> and 
##  <L> classes	respectively in which an <H> class is contained.   
##

DeclareAttribute("RClassOfHClass", IsGreensHClass); 
DeclareAttribute("LClassOfHClass", IsGreensHClass); 
DeclareAttribute("DClassOfHClass", IsGreensHClass); 
DeclareAttribute("DClassOfLClass", IsGreensLClass); 
DeclareAttribute("DClassOfRClass", IsGreensRClass); 

############################################################################
##
#A  GroupHClassOfGreensDClass( <Dclass> )
##
##  for a D class <Dclass> of a semigroup,
##  returns a group H class of the D class, or `fail' if there is no
##  group H class.
##

DeclareAttribute("GroupHClassOfGreensDClass",IsGreensDClass);

#############################################################################
##
#P  IsRegularDClass( <Dclass> )
##
##  returns `true' if the Greens D class <Dclass> is regular. A  D  class  is
##  regular if and only if each of its elements is regular, which in turn  is
##  true if and only if any one element of <Dclass> is  regular.  Idempotents
##  are regular since $eee=e$ so it follows that a Greens D class  containing
##  an idempotent is regular. Conversely, it is true that a regular  D  class
##  must contain at least one idempotent. (See~\cite{Howie76}, Prop.~3.2).
##

DeclareProperty("IsRegularDClass", IsGreensDClass);

#############################################################################
##
#P  IsGroupHClass( <Hclass> )
##
##  returns `true' if the Greens H class <Hclass> is a group, which  in  turn
##  is true if and only if <Hclass>^2 intersects <Hclass>.
##

DeclareProperty("IsGroupHClass", IsGreensHClass);

#############################################################################
##
#A  EggBoxOfDClass( <Dclass> )
##
##  returns for a Green's D class <Dclass> a matrix whose  rows  represent  R
##  classes and columns represent L classes. The entries are the H classes.
##

DeclareAttribute("EggBoxOfDClass", IsGreensDClass);

#############################################################################
##
#F  DisplayEggBoxOfDClass( <Dclass> )
##
##  displays a ``picture'' of the D class <Dclass>, as an array of 1s and 0s.
##  A 1 represents a group H class.
##

DeclareGlobalFunction("DisplayEggBoxOfDClass");

#######################
#######################

DeclareAttribute("InternalRepGreensRelation", IsGreensRelation);
DeclareAttribute("CanonicalGreensClass", IsGreensClass);
#JDM Should be IsTransformationSemigroup
DeclareOperation("DisplayEggBoxesOfSemigroup", [IsSemigroup]);

